from . channel_order import ChannelOrder
from .. colors import gamma as _gamma
from .. project import data_maker, project
import threading, time


class DriverBase(object):
    """Base driver class to build other drivers from."""

    # If set_device_brightness is not empty, it's a method that allows you
    # to directly set the brightness for the device.
    #
    # If it is empty, then the brightness is used as a scale factor in rendering
    # the pixels.
    set_device_brightness = None

    def __init__(self, num=0, width=0, height=0, c_order=ChannelOrder.RGB,
                 gamma=None, maker=data_maker.MAKER, **kwds):
        project.raise_if_unknown_attributes(kwds, 'driver', self)

        if num == 0:
            num = width * height
            if num == 0:
                raise ValueError("Either num, or width and height are needed!")

        self.make_packet, self.color_list = maker
        self.numLEDs = num
        gamma = gamma or _gamma.DEFAULT
        self.gamma = gamma

        if isinstance(c_order, str):
            c_order = ChannelOrder.make(c_order)

        self.c_order = c_order
        self.perm = ChannelOrder.ORDERS.index(c_order)

        self.pixel_positions = None

        self.width = width
        self.height = height
        self._buf = self.make_packet(self.bufByteCount())

        self.lastUpdate = 0
        self.brightness_lock = threading.Lock()
        self._brightness = 255
        self._waiting_brightness = None

    def set_pixel_positions(self, pixel_positions):
        pass

    def set_colors(self, colors, pos):
        self._colors = colors
        self._pos = pos

    def cleanup(self):
        pass

    def bufByteCount(self):
        return 3 * self.numLEDs

    def sync(self):
        """

        The sync() method is called after the entire frame has been
        sent to the device to indicate that it may now be displayed.

        This is particularly useful when there are multiple drivers comprising
        one display which all need to display the next frame at exactly the same
        time.
        """
        pass

    def _compute_packet(self):
        """Compute the packet from the colors and position.

        Eventually, this will run on the compute thread.
        """
        pass

    def _send_packet(self):
        """Send the packet to the driver.

        Eventually, this will run on an I/O thread.
        """
        pass

    def update_colors(self):
        start = time.time()

        with self.brightness_lock:
            # Swap in a new brightness.
            brightness, self._waiting_brightness = (
                self._waiting_brightness, None)

        if brightness is not None:
            self._brightness = brightness
            if self.set_device_brightness:
                self.set_device_brightness(brightness)

        self._compute_packet()
        self._send_packet()

        self.lastUpdate = time.time() - start

    def set_brightness(self, brightness):
        with self.brightness_lock:
            self._waiting_brightness = brightness

    def _render(self):
        if self.set_device_brightness:
            level = 1.0
        else:
            level = self._brightness / 255.0
        gam, (r, g, b) = self.gamma.get, self.c_order
        for i in range(self.numLEDs):
            c = [int(level * x) for x in self._colors[i + self._pos]]
            self._buf[i * 3:(i + 1) * 3] = gam(c[r]), gam(c[g]), gam(c[b])



from . import collection


class Sequence(collection.Collection):
    def step(self, amt=1):
        self.index += 1
        if self.index >= len(self.animations):
            if self.runner.until_complete:
                self.completed = True
            else:
                self.index = 0

        if not self.completed and self.animations:
            self.current_animation.run_all_frames()



import uuid
from ... util import log
from . SimpleWebSocketServer import WebSocket, SimpleWebSocketServer

Server = SimpleWebSocketServer


class Client(WebSocket):

    def __init__(self, *args, driver, pixel_positions):
        super().__init__(*args)
        self.driver = driver
        self.connected = False
        self.pixel_positions = pixel_positions
        self.oid = None
        log.debug('Server started...')

    def handleConnected(self):
        log.debug('Connected:{}'.format(self.address))
        self.connected = True
        self.oid = uuid.uuid1()
        self.driver.add_websock(self.oid, self.send_pixels)
        self.sendMessage(bytearray([0x00, 0x00]) + self.pixel_positions)

    def handleClose(self):
        self.driver.remove_websock(self.oid)
        self.connected = False
        log.debug('Closed:{}'.format(self.address))

    def handleMessage(self):
        pass

    def send_pixels(self, pixels):
        if self.connected:
            self.sendMessage(bytearray([0x00, 0x01]) + pixels)



import functools
from ...layout.geometry.rotation import Rotation

USAGE = """A Rotation is represented by a string or an integer

Possible Rotations are 0, 1, 2, 3, or 0, 90, 180, 350, or
""" + ', '.join(Rotation.__members__)


ROTATION_NAMES = {
    0: Rotation.ROTATE_0,
    1: Rotation.ROTATE_270,
    2: Rotation.ROTATE_180,
    3: Rotation.ROTATE_90,

    90: Rotation.ROTATE_90,
    180: Rotation.ROTATE_180,
    270: Rotation.ROTATE_270,
}


@functools.singledispatch
def make(c):
    raise ValueError("Don't understand type %s" % type(c), USAGE)


@make.register(str)
def _(c):
    try:
        c = int(c)
    except:
        return Rotation[c]
    else:
        return ROTATION_NAMES[c]


@make.register(int)
def _(c):
    return ROTATION_NAMES[c]



import ctypes
from multiprocessing.sharedctypes import RawArray
from .. util import log

try:
    import numpy
except:
    numpy = None

USE_NUMPY = False


def Maker(floating=None, shared_memory=False, use_numpy=USE_NUMPY):
    def list_maker(size):
        return [(0, 0, 0)] * size

    if use_numpy and not numpy:
        log.error('numpy module is not available.')
        use_numpy = False

    if use_numpy:
        log.info('Using numpy')

    if not (shared_memory or use_numpy):
        return bytearray, list_maker

    byte_type, float_type = ctypes.c_uint8, ctypes.c_float
    floating = (not shared_memory) if (floating is None) else floating
    number_type = float_type if floating else byte_type

    if shared_memory:
        def shared_list_maker(type):
            # Note https://stackoverflow.com/questions/37705974/
            return lambda size: RawArray(type, size)

        return shared_list_maker(byte_type), shared_list_maker(3 * number_type)

    def numpy_list_maker(size):
        return numpy.array(list_maker(size), dtype=number_type)

    return bytearray, numpy_list_maker


MAKER = Maker()
ColorList = MAKER[1]



import loady
from .. util import log
from .. project import project


class Runner(object):

    def __init__(self, amt=1, fps=None, sleep_time=0, max_steps=0,
                 until_complete=False, max_cycles=0, seconds=None,
                 threaded=False, main=None, **kwds):
        project.raise_if_unknown(kwds, 'attribute', 'run')

        if max_steps < 0:
            log.error('max_steps %s < 0', max_steps)
            max_steps = 0
        if sleep_time < 0:
            log.error('sleep_time %s < 0', sleep_time)
            sleep_time = 0
        if max_cycles < 0:
            log.error('max_cycles %s < 0', max_cycles)
            max_cycles = 0

        if sleep_time and fps:
            log.error('sleep_time=%s and fps=%s cannot both be set',
                      sleep_time, fps)
            sleep_time = 0
        if seconds and max_steps:
            log.error('seconds=%s and max_steps=%s cannot both be set',
                      seconds, max_steps)
            max_steps = 0

        self.amt = amt
        self.fps = fps
        self.sleep_time = 1 / fps if fps else sleep_time
        self.until_complete = until_complete
        if seconds is None:
            self.max_steps = max_steps
        else:
            self.max_steps = int(seconds / self.sleep_time)
        self.max_cycles = max_cycles
        self.seconds = seconds
        self.threaded = threaded
        self.main = main and loady.code.load(main)
